"""
Install dependencies and copies includes into a zipped file ready for deployment.

The resulting zip file is structured correctly to be deployed to the lambda
function/layer target via an S3 upload and subsequent publish command.
"""
import argparse
import typing

from reviser import bundling
from reviser import interactivity


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return ["--reinstall"]


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for this command."""
    parser.add_argument(
        "--reinstall",
        action="store_true",
        help="""
            Add this flag to reinstall dependencies on a repeated
            bundle operation. By default, dependencies will remain
            cached for the lifetime of the shell to speed up the
            bundling process. This will force dependencies to be
            installed even if they had been installed previously.
            """,
    )


def run(ex: "interactivity.Execution"):
    """Execute a bundle operation on the selected function/layer targets."""
    bundled_targets = bundling.create(
        context=ex.shell.context,
        selection=ex.shell.selection,
        reinstall=ex.args.get("reinstall", False),
    )
    print("\n\n")
    return ex.finalize(
        status="BUNDLED",
        message="Selected items have been bundled.",
        info={
            "items": [n for t in bundled_targets.targets for n in t.names],
        },
        echo=True,
    )



"""
Allow for selecting subsets of the targets within the loaded configuration.

The subsets are fuzzy-matched unless the --exact flag is used.
"""
import argparse
import dataclasses
import typing

from reviser import definitions
from reviser import interactivity


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return ["*", "--exact", "--functions", "--layers"]


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for the select command."""
    parser.add_argument(
        "--functions",
        "--function",
        "--func",
        "-f",
        dest="functions",
        action="store_true",
        help="""
        When specified, functions will be selected. This will default to true if
        neither of --functions or --layers is specified. Will default to false if
        --layers is specified.
        """,
    )
    parser.add_argument(
        "--layers",
        "--layer",
        "-l",
        dest="layers",
        action="store_true",
        help="""
        When specified, layers will be selected. This will default to true if
        neither of --functions or --layers is specified. Will default to false if
        --functions is specified.
        """,
    )
    parser.add_argument(
        "name",
        nargs="*",
        help="""
        Specifies the value to match against the function
        and layer target names available from the configuration.
        This can include shell-style wildcards and will also
        match against partial strings. If the --exact flag is
        specified, this value must exactly match one of the targets
        instead of the default fuzzy matching behavior.
        """,
    )
    parser.add_argument(
        "--exact",
        action="store_true",
        help="Forces the match to be exact instead of fuzzy.",
    )


def _update_exact_selection(
    names: typing.List[str],
    functions: bool,
    layers: bool,
    selection: "definitions.Selection",
) -> "definitions.Selection":
    """Create a new selection object updated with the specified exact match."""
    both = not functions and not layers

    layer_names = []
    if layers or both:
        layer_names = list(names)

    function_names = []
    if functions or both:
        function_names = list(names)

    return dataclasses.replace(
        selection,
        layer_names=layer_names,
        layer_needles=[],
        function_names=function_names,
        function_needles=[],
        bundle_all=False,
    )


def _update_fuzzy_selection(
    names: typing.List[str],
    functions: bool,
    layers: bool,
    selection: "definitions.Selection",
) -> "definitions.Selection":
    """Create a new selection object updated with the specified fuzzy match."""
    both = not functions and not layers

    layer_needles = []
    if layers or both:
        layer_needles = list(names)

    function_needles = []
    if functions or both:
        function_needles = list(names)

    return dataclasses.replace(
        selection,
        function_names=[],
        function_needles=function_needles,
        layer_names=[],
        layer_needles=layer_needles,
        bundle_all=False,
    )


def _get_names(ex: "interactivity.Execution") -> typing.List[str]:
    """Get selection names to use as filters."""
    names: typing.Union[str, typing.List[str]] = ex.args.get("name") or "*"
    if names and isinstance(names, str):
        return [names]

    return list(names)


def _to_names(
    targets: typing.List["definitions.Target"],
) -> typing.Union[str, typing.List[str]]:
    """Convert list of targets to a list of target names for output."""
    if not targets:
        return "None"

    return [n for t in targets for n in t.names]


def run(ex: "interactivity.Execution"):
    """Specify the target function(s) and/or layer(s) to target."""
    selection: "definitions.Selection" = ex.shell.selection
    is_exact = ex.args.get("exact", False)
    functions = ex.args.get("functions", False)
    layers = ex.args.get("layers", False)
    both = not functions and not layers
    names = _get_names(ex)

    if both and names == ["*"]:
        status = "ALL"
        message = "Selection has been cleared. All items are now selected."
        ex.shell.selection = dataclasses.replace(
            selection,
            function_needles=["*"],
            layer_needles=["*"],
            bundle_all=True,
        )
    elif is_exact:
        status = "EXACT"
        message = "Exact selection has been applied."
        ex.shell.selection = _update_exact_selection(
            names=names,
            functions=functions,
            layers=layers,
            selection=selection,
        )
    else:
        status = "MATCH"
        message = "Matching items have been selected."
        ex.shell.selection = _update_fuzzy_selection(
            names=names,
            functions=functions,
            layers=layers,
            selection=selection,
        )

    targets = ex.shell.context.get_selected_targets(ex.shell.selection)
    return ex.finalize(
        status=status,
        message=message,
        echo=True,
        info={
            "functions": _to_names(targets.function_targets),
            "layers": _to_names(targets.layer_targets),
        },
    )



"""Bundle configuration data structures and IO module."""
import dataclasses
import glob
import pathlib
import shutil
import typing

from reviser.definitions import abstracts
from reviser.definitions import configurations
from reviser.definitions import enumerations


def _get_paths(patterns: typing.Set[pathlib.Path]) -> typing.Set[pathlib.Path]:
    """Convert a set of file/directory patterns into a list of matching files."""
    raw = [
        pathlib.Path(item)
        for pattern in patterns
        for item in glob.iglob(str(pattern), recursive=True)
    ]
    files = [p for p in raw if not p.is_dir()]
    output = files + [
        child
        for p in raw
        for item in glob.iglob(f"{p}/**/*", recursive=True)
        if p.is_dir() and not (child := pathlib.Path(item)).is_dir()
    ]
    return set(output)


@dataclasses.dataclass(frozen=True)
class CopyPath:
    """Data structure for a file path copy from source to target."""

    source: pathlib.Path
    destination: pathlib.Path

    def copy(self) -> bool:
        """Copy the file from source to destination."""
        if not self.source.exists():
            return False
        self.destination.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(
            src=str(self.source.absolute()),
            dst=str(self.destination.absolute()),
        )
        return True


@dataclasses.dataclass(frozen=True)
class Bundle(abstracts.Specification):
    """
    Define the function or layer bundle that will be generated.

    That is this defines the function or layer that will be built and deployed as a
    zipped artifact to the lambda function or layer target.
    """

    target: "configurations.Target"

    @property
    def handler(self) -> typing.Optional[str]:
        """
        Get the function handler in the format `filename.function_name`.

        If not specified explicitly, it will default to the lambda function default
        value of `lambda_function.lambda_handler`.
        """
        if self.target.kind == enumerations.TargetType.LAYER:
            return None
        return self.get("handler", default="lambda_function.lambda_handler")

    @property
    def handler_filename(self) -> typing.Optional[str]:
        """Get the python filename associated with the specified handler."""
        if handler := self.handler:
            return "{}.py".format(handler.rsplit(".", 1)[0])
        return None

    @property
    def handler_function(self) -> typing.Optional[str]:
        """Get the python entrypoint function name for the specified handler."""
        if handler := self.handler:
            return handler.rsplit(".", 1)[-1]
        return None

    @property
    def omitted_packages(self) -> typing.List[str]:
        """
        List site-packages to skip with creating the bundle.

        This is especially useful when package dependencies are installed
        in both a function and layer definition and should be omitted from
        one such that they are not duplicated in the final lambda function
        environment.
        """
        omissions = self.get("omit_packages", default=self.get("omit_package"))
        if isinstance(omissions, str):
            return [omissions]

        return omissions or []

    @property
    def file_include_patterns(self) -> typing.Set[pathlib.Path]:
        """
        List file matching patterns that will be copied into the bundled artifact.

        These should be defined relative the root directory defined by the location of
        the configuration file and stored in the directory value of this object.
        """
        includes = [
            pathlib.Path(p)
            for p in self.get_as_list("includes", default=self.get_as_list("include"))
        ]
        if isinstance(includes, str):
            includes = [includes]

        add_default_package = (
            not includes and self.target.kind == enumerations.TargetType.FUNCTION
        )
        if add_default_package:
            # If the user hasn't specified any includes, look for python
            # packages in the root directory and include those by default
            # for functions. By default, layers should include nothing.
            includes = [
                pathlib.Path(item).parent.absolute()
                for item in glob.iglob(
                    str(self.directory.joinpath("*/__init__.py")),
                    recursive=True,
                )
            ]

        if filename := self.handler_filename:
            # The handler file should always be included if set.
            includes.append(self.directory.joinpath(filename))

        return set([self.directory.joinpath(item).absolute() for item in includes])

    @property
    def file_exclude_patterns(self) -> typing.Set[pathlib.Path]:
        """
        List file matching patterns that will be skipped when copying sources.

        These should be defined relative the root directory defined by the location of
        the configuration file and stored in the directory value of this object. Note
        that exclusions are applied to the inclusion results after so only need to
        specify things to remove from the inclusion list, e.g. a `tests` folder.
        """
        excludes = self.get("excludes", default=self.get("exclude")) or []
        if isinstance(excludes, str):
            excludes = [excludes]

        # Exclude Python and OS cache files by default.
        excludes += ["**/__pycache__", "**/*.pyc", "**/.DS_Store"]

        return set([self.directory.joinpath(item).absolute() for item in excludes])

    @property
    def file_package_exclude_patterns(self) -> typing.Set[pathlib.Path]:
        """
        List file matching patterns that will be skipped when bundling dependencies.

        These should be defined relative the site packages directory in which the
        dependencies are resolved. This is used to remove portions of a site package
        if they are not needed. This could be because they are already installed in
        a layer and do not need to be duplicated. Or they could be bloated elements
        of a package to remove and save on the bundle size. Either way, use with
        care as these could cause corruption in the site packages.
        """
        excludes = self.get(
            "package_excludes", default=self.get("package_exclude") or []
        )
        if isinstance(excludes, str):
            excludes = [excludes]

        return set(
            [
                self.target.site_packages_directory.joinpath(item).resolve()
                for item in excludes
            ]
        )

    def get_include_paths(
        self,
        relative: bool = False,
    ) -> typing.Set[pathlib.Path]:
        """Generate a list of file paths matching the bundle includes."""
        items = _get_paths(self.file_include_patterns)
        if filename := self.handler_filename:
            items.add(self.directory.joinpath(filename))
        if relative:
            return set([item.relative_to(self.directory) for item in items])
        return items

    def get_exclude_paths(
        self,
        relative: bool = False,
    ) -> typing.Set[pathlib.Path]:
        """Generate a list of file paths matching the bundle excludes."""
        items = _get_paths(self.file_exclude_patterns)
        if relative:
            return set([item.relative_to(self.directory) for item in items])
        return items

    def get_paths(
        self,
        relative: bool = False,
    ) -> typing.Set[pathlib.Path]:
        """Generate a list of file paths matching the bundle excludes."""
        return self.get_include_paths(relative=relative) - self.get_exclude_paths(
            relative=relative
        )

    def get_copy_paths(self) -> typing.List["CopyPath"]:
        """
        List paths to copy where each maps source to destination for bundling.

        Each item returned contains an included (and not excluded) mapping from the
        source paths to their respective destination path in the bundle directory.
        """
        return [
            CopyPath(
                source=self.directory.joinpath(p),
                destination=self.target.bundle_directory.joinpath(p),
            )
            for p in self.get_paths(relative=True)
        ]

    def get_site_package_paths(self) -> typing.Set[pathlib.Path]:
        """
        List paths in the target bundle site packages directory to include in bundles.

        Omitted packages will be excluded from this list if any are specified in
        the bundle. Also, user-defined package_excludes paths will also be omitted.
        """
        directory = self.target.site_packages_directory
        exclusions = []
        for package in self.omitted_packages:
            exclusions += [
                directory.joinpath(package, "**", "*"),
                directory.joinpath(f"{package}.*"),
                directory.joinpath(f"{package}-*"),
                directory.joinpath(f"{package}-*", "**", "*"),
            ]
        omitted_paths = _get_paths(
            set(exclusions).union(self.file_package_exclude_patterns)
        )

        paths = _get_paths({directory.joinpath("**", "*")})
        return paths - omitted_paths

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        return {
            "handler": self.handler,
            "handler_filename": self.handler_filename,
            "handler_function": self.handler_function,
            "omitted_packages": self.omitted_packages,
            "file_include_patterns": [str(p) for p in self.file_include_patterns],
            "file_exclude_patterns": [str(p) for p in self.file_exclude_patterns],
        }



"""Publisher functionality module."""
import typing

from reviser import definitions
from ..deploying import updater


def publish_function(
    target: "definitions.Target",
    s3_keys: typing.List[str],
    published_layers: typing.List["definitions.PublishedLayer"],
    description: str = None,
    dry_run: bool = False,
):
    """Publish updated versions of the functions after upload."""
    client = target.client("lambda")
    for name, key in zip(target.names, s3_keys):
        print(f"[PUBLISHING]: Deploying code bundle to {name} $LATEST")
        response = None

        if not dry_run:
            response = client.update_function_code(
                FunctionName=name,
                S3Bucket=target.bucket,
                S3Key=key,
                Publish=False,
            )

        updater.update_function_configuration(
            function_name=name,
            target=target,
            published_layers=published_layers,
            dry_run=dry_run,
        )

        print("[PUBLISHING]: Publishing new version from bundle")
        if response and not dry_run:
            response = client.publish_version(
                FunctionName=response["FunctionName"],
                CodeSha256=response["CodeSha256"],
                Description=description or "",
            )
            print(
                "[PUBLISHED]: Function {} ({})".format(
                    name,
                    response["Version"],
                )
            )
            print("  - Version:", response["FunctionArn"])
            print("  - Runtime:", response["Runtime"])

    print("[DEPLOYED]: Lambda function code has been deployed\n")


def publish_layer(
    target: "definitions.Target",
    s3_keys: typing.List[str],
    description: str = None,
    dry_run: bool = False,
) -> typing.List["definitions.PublishedLayer"]:
    """Publish an updated version of the layer after bundle has been uploaded to S3."""
    published_layers = []
    client = target.client("lambda")

    for name, key in zip(target.names, s3_keys):
        print(f"[PUBLISHING]: Publishing code bundle to {name} layer")
        if dry_run:
            continue

        response = client.publish_layer_version(
            LayerName=name,
            Description=description or "",
            Content={
                "S3Bucket": target.bucket,
                "S3Key": key,
            },
            CompatibleRuntimes=[f"python{definitions.RUNTIME_VERSION}"],
        )
        print("[PUBLISHED]: Layer {} ({})".format(name, response["Version"]))
        print("  - Layer:", response["LayerArn"])
        print("  - Version:", response["LayerVersionArn"])
        published_layers.append(definitions.PublishedLayer(response=response))

    print("[DEPLOYED]: Lambda layer code has been deployed\n")
    return published_layers



"""Definitions and IO for AWS connection data."""
import typing

import boto3
from botocore.client import BaseClient
from botocore.credentials import Credentials


class AwsConnection:
    """Boto3 Session wrapper that facilitates broader session management."""

    def __init__(self, session: boto3.Session = None):
        """Create an AwsConnection object from the boto3 session or a default one."""
        self._session = session or boto3.Session()
        self._aws_account = self.client("sts").get_caller_identity()
        self._aws_account_aliases = self.client("iam").list_account_aliases()

    @property
    def session(self) -> boto3.Session:
        """Get the session object associated with this connection."""
        return self._session

    @property
    def user_id(self) -> str:
        """Get the fully-qualified user ID for the associated session."""
        return self._aws_account.get("UserId")

    @property
    def user_arn(self) -> str:
        """Get the ARN of the user associated with this session."""
        return self._aws_account.get("Arn")

    @property
    def user_slug(self) -> str:
        """
        Get the display slug identifying the user.

        This is derived from the user ARN containing account ID and user name or
        assumed role for use in shell prompt.
        """
        if "iam::" in self.user_arn:
            return self.user_arn.split("iam::")[-1]
        return self.user_arn.split("sts::")[-1].rsplit("/", 1)[0]

    @property
    def aws_account_id(self) -> str:
        """Get the AWS account ID for the associated session."""
        return self._aws_account.get("Account")

    @property
    def aws_account_alias(self) -> typing.Optional[str]:
        """Get the AWS account alias associated with the session."""
        aliases = self._aws_account_aliases.get("AccountAliases")
        if aliases:
            return aliases[0]
        return None

    def get_credentials(self) -> Credentials:
        """Get the credentials associated with the session."""
        return self.session.get_credentials()

    def client(self, *args, **kwargs) -> BaseClient:
        """Create a new client of the specified type."""
        return self._session.client(*args, **kwargs)



"""Update functionality module."""
import typing
import yaml

from botocore.client import BaseClient

from reviser import definitions
from reviser import servicer


def _filter_layer_updates(
    function_name: str,
    target: "definitions.Target",
    published_layers: typing.List["definitions.PublishedLayer"],
) -> typing.List[str]:
    """Filter layer updates to ones attachable to the specified function."""
    updates = {p.arn: p.versioned_arn for p in published_layers}
    return [
        updates.get(a.arn) or a.arn
        for a in target.layer_attachments
        if a.is_attachable(function_name)
    ]


def _to_versioned_arn(client: BaseClient, arn: str) -> str:
    """
    Convert the ARN into a versioned ARN if not already.

    Return an empty string if no versioned ARN exists.
    """
    if arn and arn.count(":") == 7:
        # Having seven ":" in a layer arn means that the version is included.
        # An unversioned arn would only have six ":".
        return arn

    if arn and (found_versions := servicer.get_layer_versions(client, arn)):
        version = found_versions[-1].version
        return f"{arn}:{version}"

    print(f"\n[WARNING]: No versions found for layer:\n   {arn}\n")
    return ""


def _get_layer_updates(
    client: BaseClient,
    function_name: str,
    target: "definitions.Target",
    current_configuration: dict,
    published_layers: typing.List["definitions.PublishedLayer"],
) -> typing.Optional[typing.List[str]]:
    """
    Create an updated list of layer ARNs to update on the function.

    These will be layer configurations to apply during the function configuration
    update now that new layers may have been made available.
    """
    if target.ignores_any("layer", "layers"):
        return None

    existing = [item["Arn"] for item in (current_configuration.get("Layers") or [])]
    layers = _filter_layer_updates(function_name, target, published_layers)

    latest = [
        versioned_arn
        for arn in layers
        if (versioned_arn := _to_versioned_arn(client, arn))
    ]

    if latest == existing:
        return None

    return latest


def _get_handler_update(
    target: "definitions.Target",
    current_configuration: dict,
) -> typing.Optional[str]:
    """Specify an updated handler if a change is found."""
    latest = target.bundle.handler
    existing = current_configuration.get("Handler")
    if target.ignores_any("handler") or latest == existing:
        return None
    return latest


def _get_runtime_update(
    target: "definitions.Target",
    current_configuration: dict,
) -> typing.Optional[str]:
    """Specify an updated runtime if a change is found."""
    latest = f"python{definitions.RUNTIME_VERSION}"
    existing = current_configuration.get("Runtime")
    if target.ignores_any("runtime") or latest == existing:
        return None
    return latest


def _get_memory_update(
    target: "definitions.Target",
    current_configuration: dict,
) -> typing.Optional[int]:
    """Specify an updated memory value if a change is found."""
    existing = current_configuration.get("MemorySize")
    latest = target.memory
    if target.ignores_any("memory") or latest == existing:
        return None
    return latest


def _get_timeout_update(
    target: "definitions.Target",
    current_configuration: dict,
) -> typing.Optional[int]:
    """Specify an updated timeout value if a change is found."""
    existing = current_configuration.get("Timeout")
    latest = target.timeout
    if target.ignores_any("timeout") or latest == existing:
        return None
    return latest


def _get_variable_updates(
    function_name: str,
    target: "definitions.Target",
    current_configuration: dict,
) -> typing.Optional[dict]:
    """Determine the environment variable updates for the function."""
    if target.ignores_any("variable", "variables"):
        return None

    existing = current_configuration.get("Environment", {}).get("Variables", {})

    def _get_value(variable: "definitions.EnvironmentVariable"):
        """Fetch the environment variable value based on its settings."""
        if variable.preserve:
            return existing.get(variable.name)
        return variable.get_value(function_name)

    latest = {e.name: v for e in target.variables if (v := _get_value(e))}
    if latest == existing:
        return None
    return {"Variables": latest}


def update_function_configuration(
    function_name: str,
    target: "definitions.Target",
    published_layers: typing.List["definitions.PublishedLayer"],
    dry_run: bool = False,
):
    """Update the function configuration as part of a deployment process."""
    client = target.client("lambda")
    current = client.get_function_configuration(FunctionName=function_name)

    changes = dict(
        Layers=_get_layer_updates(
            client,
            function_name,
            target,
            current_configuration=current,
            published_layers=published_layers,
        ),
        Runtime=_get_runtime_update(target, current),
        MemorySize=_get_memory_update(target, current),
        Timeout=_get_timeout_update(target, current),
        Environment=_get_variable_updates(function_name, target, current),
        Handler=_get_handler_update(target, current),
    )

    modifications = {k: v for k, v in changes.items() if v is not None}
    if not modifications:
        print(f"[UNCHANGED]: {function_name} configuration unchanged.")
        return

    print(f"[MODIFYING]: Updating {function_name} configuration values:")
    print(yaml.safe_dump(modifications))

    if dry_run:
        print("[DRY_RUN]: No configuration updates applied.")
        return

    client.update_function_configuration(FunctionName=function_name, **modifications)



"""
Display the configs merged from its source file, dynamic values and defaults.

Use this to inspect and validate that the loaded configuration meets expectations when
parsed into the reviser shell.
"""
import argparse
import typing

import yaml

from reviser import interactivity


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return []


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for this command."""
    pass


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Execute a bundle operation on the selected function/layer targets."""
    print(yaml.safe_dump(ex.shell.context.configuration.serialize()))
    return ex.finalize(status="SUCCESS", message="Configuration displayed.")



"""Remove old function and/or layer versions for the selected targets."""
import argparse
import textwrap
import typing

from botocore.client import BaseClient

from reviser import definitions
from reviser import interactivity
from reviser import servicer


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return ["--start", "--end", "--dry-run"]


def populate_subparser(parser: argparse.ArgumentParser):
    """Add subcommand options for this command."""
    parser.add_argument(
        "--start",
        type=int,
        default=None,
        help="""
            Keep versions lower (earlier/before) this one. A negative value can be
            specified for relative indexing in the same fashion as Python lists.
            """,
    )
    parser.add_argument(
        "--end",
        type=int,
        default=None,
        help="""
            Do not prune versions higher than this value. A negative value can be
            specified for relative indexing in the same fashion as Python lists.
            """,
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Echo pruning operation without actually executing it.",
    )
    parser.add_argument(
        "-y",
        "--yes",
        action="store_true",
        help="Run the prune process without reviewing first.",
    )


def _resolve_version(
    versions: typing.Union[
        typing.List[definitions.LambdaLayer],
        typing.List[definitions.LambdaFunction],
    ],
    value: int = None,
) -> typing.Optional[int]:
    """
    Normalize specified versions by making them absolute, positive version integers.

    Leaves None and positive integer values alone, but will convert relative negative
    integers into their positive integer equivalents based on the available versions
    included in the version argument.

    :param versions:
        List of function or layer versions that currently exist, which will be used
        to resolve relative, negative values.
    :param value:
        Value to resolve.
    """
    if value is None or value >= 0:
        return value

    highest_version = max(
        [int(v.version or 0) for v in versions if "$" not in str(v.version)]
    )
    return max(0, highest_version + value)


def _is_within_range(
    function: "definitions.LambdaFunction",
    start_value: typing.Optional[int],
    end_value: typing.Optional[int],
) -> bool:
    """Determine if the function version is within the start and end range."""
    return (start_value is None or start_value <= int(function.version or 0)) and (
        end_value is None or int(function.version or 0) <= end_value
    )


def _get_function_removal_versions(
    lambda_client: BaseClient,
    function_name: str,
    start: int = None,
    end: int = None,
):
    """Get ARNs for the function that should be pruned."""
    versions = servicer.get_function_versions(lambda_client, function_name)
    start_value = _resolve_version(versions, start)
    end_value = _resolve_version(versions, end)
    return [
        v.arn
        for v in versions
        if v.version != "$LATEST"
        and v.arn
        and _is_within_range(v, start_value, end_value)
        and not v.aliases
    ]


def _prune_function(
    lambda_client: BaseClient,
    function_name: str,
    start: int = None,
    end: int = None,
    dry_run: typing.Optional[bool] = False,
    confirm: typing.Optional[bool] = True,
):
    """
    Execute a pruning operation on the given lambda functions.

    :param lambda_client:
        Boto3 client interface for the lambda service
    :param function_name:
        Name of the function to prune
    :param start:
        First version of the lambda function to prune. Versions below
        that will be retained.
    :param end:
        Inclusive end version of the lambda function to prune. Versions
        above that will be retained.
    :param dry_run:
        If true, only print out what would happen and don't actually prune
        the versions.
    :param confirm:
        Whether or not to ask before proceeding with the prune operation.
    """
    removal_arns = _get_function_removal_versions(
        lambda_client=lambda_client,
        function_name=function_name,
        start=start,
        end=end,
    )

    print("\nARN Versions to be removed:")
    print(
        textwrap.indent(
            "\n".join(removal_arns),
            prefix="  - ",
        )
    )

    if dry_run:
        print("\n[DRY RUN]: Skipped removal process.")
        return

    abort = confirm and not (
        input("\nExecute prune action [y/N]? ") or ""
    ).lower().startswith("y")
    if abort:
        print("\n[ABORTED]: Skipped removal process.")
        return

    for arn in removal_arns:
        servicer.remove_function_version(lambda_client, arn)

    return removal_arns


def _get_layer_removal_versions(
    lambda_client: BaseClient,
    layer_name: str,
    start: int = None,
    end: int = None,
):
    """Get ARNs for the layer that should be pruned."""
    versions = servicer.get_layer_versions(lambda_client, layer_name)
    start_value = _resolve_version(versions, start)
    end_value = _resolve_version(versions, end)
    removals = [
        version
        for version in versions[:-1]
        if (start_value is None or start_value <= int(version.version or 0))
        and (end_value is None or int(version.version or 0) <= end_value)
    ]
    return [r.arn for r in removals if r.arn]


def _prune_layer(
    lambda_client: BaseClient,
    layer_name: str,
    start: int = None,
    end: int = None,
    dry_run: typing.Optional[bool] = False,
    confirm: typing.Optional[bool] = True,
):
    """
    Execute a pruning operation on the given lambda layers.

    :param lambda_client:
        Boto3 client interface for the lambda service.
    :param layer_name:
        Name of the layer to prune
    :param start:
        First version of the lambda layer to prune. Versions below
        that will be retained.
    :param end:
        Inclusive end version of the lambda layer to prune. Versions
        above that will be retained.
    :param dry_run:
        If true, only print out what would happen and don't actually prune
        the versions.
    :param confirm:
        Whether or not to ask before proceeding with the prune operation.
    """
    removal_arns = _get_layer_removal_versions(
        lambda_client=lambda_client,
        layer_name=layer_name,
        start=start,
        end=end,
    )

    print("\nARN Versions to be removed:")
    print(
        textwrap.indent(
            "\n".join(removal_arns or []),
            prefix="  - ",
        )
    )

    if dry_run:
        print("\n[DRY RUN]: Skipped removal process.")
        return

    abort = confirm and not (
        input("\nExecute prune action [y/N]? ") or ""
    ).lower().startswith("y")
    if abort:
        print("\n[ABORTED]: Skipped removal process.")
        return

    for arn in removal_arns:
        print(f"[PRUNING]: {arn}")
        servicer.remove_layer_version(lambda_client, arn)

    return removal_arns


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Execute the pruning operation on the selected or specified targets."""
    selected = ex.shell.context.get_selected_targets(ex.shell.selection)
    targets = sorted(selected.targets, key=lambda t: t.kind.value)

    caller = {
        definitions.TargetType.FUNCTION: _prune_function,
        definitions.TargetType.LAYER: _prune_layer,
    }

    results = {
        name: caller[target.kind](
            target.client("lambda"),
            name,
            start=ex.args.get("start"),
            end=ex.args.get("end"),
            dry_run=ex.args.get("dry_run") or False,
            confirm=not ex.args.get("yes") or False,
        )
        for target in targets
        for name in target.names
    }

    return ex.finalize(
        status="PRUNED",
        message="Specified versions have been removed.",
        info={name: removed_arns for name, removed_arns in results.items()},
        echo=True,
    )



"""Dependency configuration data structures and IO module."""
import dataclasses
import json
import pathlib
import subprocess
import sys
import typing

from reviser import utils
from reviser.definitions import abstracts
from reviser.definitions import configurations
from reviser.definitions import enumerations


@dataclasses.dataclass(frozen=True)
class Dependency(abstracts.Specification):
    """Package dependency data structure."""

    target: "configurations.Target"

    @property
    def kind(self) -> "enumerations.DependencyType":
        """Get the kind of dependency."""
        return enumerations.DependencyType(
            value=self.get(
                "kind",
                default=enumerations.DependencyType.PIP.value,
            )
        )

    @property
    def file(self) -> typing.Optional[pathlib.Path]:
        """Get the optionally-specified path used to load dependencies."""
        if value := self.get("file"):
            return self.directory.joinpath(value).absolute()

        if self.get("packages") is not None:
            return None

        default = self.directory.joinpath(
            enumerations.DefaultFile.from_dependency_type(self.kind).value
        )
        return default if default.exists() else None

    @property
    def packages(self) -> typing.List[str]:
        """Get the explicitly defined packages."""
        return self.get_first_as_list(["packages"], ["package"], default=[])

    def get_package_names(self) -> typing.List[str]:
        """List package names to install from the various package sources."""
        return []


@dataclasses.dataclass(frozen=True)
class PipDependency(Dependency):
    """Pip package dependency data structure."""

    def get_package_names(self) -> typing.List[str]:
        """List of package names to install collected from the various sources."""
        packages = self.packages.copy()
        if self.file:
            packages += [
                item
                for line in (self.file.read_text() or "").split("\n")
                if (item := line.strip())
            ]
        return packages

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        return {
            "kind": self.kind.value,
            "packages": self.get_package_names(),
        }


@dataclasses.dataclass(frozen=True)
class PipperDependency(Dependency):
    """Pipper package dependency data structure."""

    @property
    def prefix(self) -> typing.Optional[str]:
        """Get the custom S3 key prefix in the bucket where the repository resides."""
        return self.get("prefix")

    @property
    def bucket(self) -> typing.Optional[str]:
        """Get the S3 bucket where the pipper repository resides."""
        return utils.get_matching_bucket(
            buckets=self.get_first(["buckets"], ["bucket"]),
            aws_region=self.target.aws_region,
            aws_account_id=self.connection.aws_account_id,
        )

    def get_package_data(self) -> typing.Optional[dict]:
        """Get the  data stored in the specified package file."""
        if self.file:
            return json.loads(self.file.read_text())
        return None

    def get_package_names(self) -> typing.List[str]:
        """List collected package names to install from various definition sources."""
        packages = self.packages.copy()
        if data := self.get_package_data():
            packages += data.get("dependencies") or []
        return packages

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        return {
            "kind": self.kind.value,
            "prefix": self.prefix,
            "bucket": self.bucket,
            "packages": self.get_package_names(),
        }


def _find_poetry_executable() -> str:
    """Find the poetry executable path, or raise an error if not found."""
    directories = [
        pathlib.Path(sys.prefix).joinpath("bin"),
        pathlib.Path(sys.executable).parent,
    ]
    directories = [d for d in directories if d.exists()]

    finder = (
        str(p)
        for d in directories
        for p in d.iterdir()
        if p.name == "poetry" or p.name.startswith("poetry.")
    )
    try:
        return next(finder)
    except StopIteration:
        raise RuntimeError(
            "Unable to find poetry installation in current Python environment."
        )


@dataclasses.dataclass(frozen=True)
class PoetryDependency(Dependency):
    """Poetry package dependency data structure."""

    @property
    def extras(self) -> typing.List[str]:
        """List extra packages to install."""
        return self.get_as_list("extras", default=self.get_as_list("extra")) or []

    def get_package_names(self) -> typing.List[str]:
        """List collected package names to install from various definition sources."""
        packages = self.packages.copy()
        if not self.file:
            return packages

        executable = _find_poetry_executable()
        command = [
            executable,
            "export",
            "--format=requirements.txt",
            "--without-hashes",
        ]

        for group in self.extras:
            command.append(f"--extras={group}")

        result = subprocess.run(command, stdout=subprocess.PIPE, check=True)

        packages += [
            item
            for line in result.stdout.decode().strip().split("\n")
            if (item := line.split(";")[0].strip())
        ]

        return packages

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        return {
            "kind": self.kind.value,
            "packages": self.get_package_names(),
        }



"""Data structure and IO module for environment variables definitions."""
import dataclasses
import fnmatch
import typing

from reviser.definitions import abstracts
from reviser.definitions import configurations


@dataclasses.dataclass(frozen=True)
class EnvironmentVariable(abstracts.Specification):
    """Data structure for function environment variables."""

    target: "configurations.Target"

    @property
    def name(self) -> str:
        """Get the name for the environment variable."""
        if (name := self.get("name")) is not None:
            return name

        if (arg := self.get("arg")) is not None:
            return arg.split("=")[0]

        return "unknown-environment-variable"

    @property
    def preserve(self) -> bool:
        """
        Specify whether or not the environment variable should be preserved.

        Preserved environment variables will not be changed during updates. Useful for
        variables that are managed or assigned by other systems.
        """
        return self.get("preserve", default=False)

    @property
    def restrictions(self) -> typing.List[str]:
        """List lambda functions to attach this environment variable to."""
        value = self.get("only", default=[])
        if isinstance(value, str):
            return [value]
        return value

    @property
    def exclusions(self) -> typing.List[str]:
        """List lambda functions not to attach this environment variable to."""
        value = self.get("except", default=[])
        if isinstance(value, str):
            return [value]
        return value

    def _get_raw_value_for_function(self, function_name: str) -> typing.Optional[str]:
        """Get raw environment variable for the specified function name."""
        if not self.has("value") and (arg := self.get("arg")):
            value = arg.split("=", 1)[1]
        else:
            value = self.get("value")

        if isinstance(value, dict):
            selector = (
                v
                for pattern, v in value.items()
                if fnmatch.fnmatch(function_name, pattern)
            )
            value = next(selector, None)

        return value

    def _get_is_included(self, function_name: str) -> bool:
        """
        Determine if the env var is included in the specified function.

        This will return true for cases where no inclusion restrictions have been
        applied. Otherwise, it will return True only if the function name matches one
        of the inclusion restriction patterns.
        """
        includer = (
            True
            for pattern in self.restrictions
            if fnmatch.fnmatch(function_name, pattern)
        )
        return next(includer, not self.restrictions)

    def _get_is_excluded(self, function_name: str) -> bool:
        """Determine if the env var is excluded from the specified function."""
        excluder = (True for e in self.exclusions if fnmatch.fnmatch(function_name, e))
        return next(excluder, False)

    def get_value(self, function_name: str) -> typing.Optional[str]:
        """Retrieve the environment variable value for the given function name."""
        is_applicable = self._get_is_included(
            function_name
        ) and not self._get_is_excluded(function_name)
        if is_applicable:
            return self._get_raw_value_for_function(function_name)
        return None

    def serialize(self) -> dict:
        """Serialize for logging and debugging purposes."""
        if self.preserve:
            kwargs = {}
        else:
            kwargs = {
                "values": {
                    n: v
                    for n in self.target.names
                    if (v := self.get_value(n)) is not None
                },
            }

        return {
            "name": self.name,
            "only": self.restrictions,
            "except": self.exclusions,
            "preserve": self.preserve,
            **kwargs,
        }



"""Configurations definitions package containing data structures and IO for configs."""
import dataclasses
import typing

from reviser import utils
from reviser.definitions import abstracts
from reviser.definitions import enumerations
from .bundling import Bundle  # noqa: F401
from .depending import Dependency  # noqa: F401
from .depending import PipDependency  # noqa: F401
from .depending import PipperDependency  # noqa: F401
from .depending import PoetryDependency  # noqa: F401
from .enviromentals import EnvironmentVariable  # noqa: F401
from .layering import AttachedLayer  # noqa: F401
from .targeting import Target  # noqa: F401


@dataclasses.dataclass(frozen=True)
class Configuration(abstracts.Specification):
    """Configuration Settings data structure."""

    @property
    def targets(self) -> typing.List["Target"]:
        """Get the duild and deploy targets."""
        return [
            Target(
                directory=self.directory,
                data=item,
                connection=self.connection,
                configuration=self,
            )
            for item in self.get("targets", default=[])
        ]

    @property
    def function_targets(self) -> typing.List["Target"]:
        """Get the build and deploy function targets."""
        return [t for t in self.targets if t.kind == enumerations.TargetType.FUNCTION]

    @property
    def layer_targets(self) -> typing.List["Target"]:
        """Get the build and deploy layer targets."""
        return [t for t in self.targets if t.kind == enumerations.TargetType.LAYER]

    @property
    def bucket(self) -> typing.Optional[str]:
        """Retrieve the bucket to use for uploading to S3."""
        return utils.get_matching_bucket(
            buckets=self.get_first(["buckets"], ["bucket"]),
            aws_region=self.aws_region,
            aws_account_id=self.connection.aws_account_id,
        )

    @property
    def aws_region(self) -> str:
        """Get the name of the region that this configuration is targeting."""
        return self.get("region") or self.connection.session.region_name or "us-east-1"

    def get_function(self, name: str) -> typing.Optional["Target"]:
        """Retrieve the target with the matching name."""
        return next((t for t in self.function_targets if name in t.names), None)

    def get_layer(self, name: str) -> typing.Optional["Target"]:
        """Retrieve the target with the matching name."""
        return next((t for t in self.layer_targets if name in t.names), None)

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        return {
            "bucket": self.bucket,
            "run": self.get("run"),
            "region": self.aws_region,
            "function_targets": [t.serialize() for t in self.function_targets],
            "layer_targets": [t.serialize() for t in self.layer_targets],
        }



"""Exit the shell and returns to the parent terminal."""
import argparse
import typing

from reviser import interactivity


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return []


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for this command."""
    pass


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Exit the shell."""
    ex.shell.shutdown = True
    return ex.finalize(
        status="EXIT",
        message="Shutting down the shell.",
        echo=True,
    )



"""
Assign an alias to the specified version of the selected or specified lambda function.

Or it will create a new alias and assign it to the specified version if the --create
flag is included. To assign an existing `test` alias to version 42 of the selected
function, the command would be:

```
> alias test 42
```

If multiple functions are currently selected, use `--function=<NAME>`
to identify the function to which the alias change will be applied.
"""
import typing
from argparse import ArgumentParser

from reviser import definitions
from reviser import interactivity
from reviser import servicer


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return ["--function", "--yes", "--create"]


def populate_subparser(parser: ArgumentParser):
    """Add alias subcommand to supplied parser."""
    parser.add_argument(
        "alias",
        help="""
        Name of an existing alias to move to the specified version,
        or the name of an alias to create and assign to the specified
        function version if the --create flag is included to allow for
        creating a new alias.
        """,
    )
    parser.add_argument(
        "version",
        default="$LATEST",
        help="""
        Version of the function that the alias should be assigned
        to. This will either be an integer value or $LATEST. To see
        what versions are available for a given function use the list
        command.
        """,
    )
    parser.add_argument(
        "--function",
        help="""
        The alias command only acts on one function. This can be
        achieved either by selecting the function target via the
        select command, or specifying the function name to apply this
        change to with this flag.
        """,
    )
    parser.add_argument(
        "--yes",
        action="store_true",
        help="""
        By default this command will require input confirmation
        before carrying out the change. Specify this flag to
        skip input confirmation and proceed without a breaking prompt.
        """,
    )
    parser.add_argument(
        "--create",
        action="store_true",
        help="""
        When specified the alias will be created instead of reassigned.
        Use this to create and assign new aliases to a function. When
        this flag is not specified, the command will fail if the alias
        doesn't exist, which helps prevent accidental alias creation.
        """,
    )


def _get_version(target: "definitions.Target", version: str) -> str:
    """Convert the version value into an absolute function version."""
    try:
        number = int(version)
    except ValueError:
        return version

    if number > 0:
        return version

    client = target.client("lambda")
    versions = servicer.get_function_versions(client, target.names[0])
    return versions[number - 1].version or "unknown"


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Execute the command invocation."""
    selected = ex.shell.context.get_selected_targets(ex.shell.selection)
    is_ambiguous = not ex.args.get("function") and (
        0 < len(selected.function_targets) > 1
        or len(selected.function_targets[0].names) > 1
    )
    if is_ambiguous:
        return ex.finalize(
            status="ERROR",
            message="""
                Ambiguous alias assignment. Only one function can be assigned
                at a time. Either modify the selection or use the --function
                flag to specify the target function for the alias change.
                """,
            echo=True,
        )

    target = selected.function_targets[0]
    name = ex.args.get("function") or target.names[0]
    client = target.client("lambda")
    alias = ex.args.get("alias")
    version = _get_version(target, ex.args.get("version", "unknown"))
    create = ex.args.get("create")
    yes = ex.args.get("yes")

    request = dict(
        FunctionName=name,
        Name=alias,
        FunctionVersion=version,
    )

    prefix = f"Create {alias} at" if create else f"Move {alias} to"
    message = f"\n{prefix} {name}:{version} [y/N]? "
    if not yes and not (input(message) or "").strip().lower().startswith("y"):
        return ex.finalize(
            status="ABORTED",
            message="No alias changes made.",
            echo=True,
        )

    if create:
        client.create_alias(**request)
    else:
        client.update_alias(**request)

    return ex.finalize(
        status="SUCCESS",
        message="Alias change has been applied.",
        info={"function": name, "alias": alias, "version": version},
        echo=True,
    )



"""Auto-completion functionality module for the shell."""
import re
import shlex
import typing

from prompt_toolkit import completion
from prompt_toolkit.document import Document

from reviser import commands
from reviser.interactivity import shells


class ShellCompleter(completion.Completer):
    """
    Implement a concrete Prompt Toolkit Completer class.

    This class is specific to the behaviors of the lambda deployer interactive shell.
    """

    def __init__(self, shell: "shells.Shell"):
        """Initialize the prompt toolkit custom auto-completer."""
        super(ShellCompleter, self).__init__()
        self.shell = shell
        self.document: typing.Optional[Document] = None
        self.event: typing.Optional[completion.CompleteEvent] = None
        self.args: typing.List[str] = []

    @property
    def action(self) -> str:
        """Get the currently specified root shell command action."""
        return self.args[0] if self.args else ""

    @property
    def last(self) -> str:
        """Get the last argument in the current line."""
        return self.args[-1] if self.args else ""

    def get_completions(
        self, document: Document, complete_event: completion.CompleteEvent
    ) -> typing.Generator[completion.Completion, None, None]:
        """
        Get the auto-completions for the current shell command state.

        Implements the completion process where the Prompt Toolkit document
        and complete event are used to determine the applicable completions
        for the current shell input state.
        """
        try:
            self.args = shlex.split(document.text.strip())
        except ValueError:
            return

        self.document = document
        self.event = complete_event

        for c in _get_completions(self):
            yield c


def _get_completions(
    completer: ShellCompleter,
) -> typing.List[completion.Completion]:
    """List completions to return by the ShellCompleter.get_completions function."""
    if not completer.document:
        return []

    if completer.document.text.lstrip().find(" ") == -1:
        return _get_action_completions(completer)

    return _get_command_completions(completer) + _get_state_completions(completer)


def _get_action_completions(completer: ShellCompleter):
    """Get matching command actions for the current shell prompt state."""
    actions = list(commands.ALIASES.keys()) + list(commands.COMMANDS.keys())
    index = len(completer.action)
    matches = [a for a in actions if a.startswith(completer.action)]
    return [completion.Completion(m, start_position=-index) for m in matches]


def _get_command_completions(completer: ShellCompleter):
    """
    Get completions specific to the currently specified shell command/subcommand.

    These completions are defined in the completer.configs module in the
    configs.COMMAND_COMPLETES dictionary.
    """
    command_module = commands.get_module(completer.action)
    get_completions = getattr(command_module, "get_completions", lambda x: [])
    options = get_completions(completer)
    index = len(completer.last)
    return [
        completion.Completion(item, start_position=-index)
        for item in options
        if item.startswith(completer.last)
    ]


def _get_state_completions(completer: ShellCompleter):
    """
    Get completion for the current state of the shell.

    At this point, this only includes lambda function/layer name completion.
    """
    index = len(completer.last)
    targets = [
        *completer.shell.context.configuration.function_targets,
        *completer.shell.context.configuration.layer_targets,
    ]
    names = [n for t in targets for n in t.names]

    pattern = re.compile(r"[\s_]+")
    names += list(
        {
            word
            for name in names
            for word in pattern.sub("-", name).split("-")
            if len(word) > 2
        }
    )

    return [
        completion.Completion(n, start_position=-index)
        for n in names
        if n.startswith(completer.last)
    ]
